@@ -9,13 +9,14 @@ use nextest_metadata::NextestExitCode;
 use nextest_runner::{
     config::core::{ConfigExperimental, ToolName},
     errors::{
-        CacheDirError, RecordReadError, RecordSetupError, RunIdResolutionError, RunStoreError,
-        TestListFromSummaryError, UserConfigError, *,
+        CacheDirError, PortableRecordingError, PortableRecordingReadError, RecordReadError,
+        RecordSetupError, RunIdResolutionError, RunStoreError, TestListFromSummaryError,
+        UserConfigError, *,
     },
     helpers::{format_interceptor_too_many_tests, plural},
     indenter::DisplayIndented,
     list::OwnedTestInstanceId,
-    record::{ReplayabilityStatus, RunListAlignment},
+    record::{ReplayabilityStatus, RunListAlignment, StoreVersionIncompatibility},
     redact::Redactor,
     run_mode::NextestRunMode,
     runner::{DebuggerCommand, TracerCommand},
@@ -416,6 +417,8 @@ pub enum ExpectedError {
         #[source]
         err: std::io::Error,
     },
+    #[error("error extracting from portable recording: {message}")]
+    DebugExtractRecordingError { message: String },
     #[error("run ID resolution error")]
     RunIdResolutionError {
         #[source]
@@ -426,13 +429,20 @@ pub enum ExpectedError {
         #[source]
         err: RecordReadError,
     },
-    #[error(
-        "run {run_id} has unsupported store format version {found} (this nextest supports version {supported})"
-    )]
-    UnsupportedStoreFormatVersion {
+    #[error("run {run_id} has incompatible store format: {incompatibility}")]
+    StoreVersionIncompatible {
         run_id: ReportUuid,
-        found: u32,
-        supported: u32,
+        incompatibility: StoreVersionIncompatibility,
+    },
+    #[error("error creating portable recording")]
+    PortableRecordingError {
+        #[source]
+        err: PortableRecordingError,
+    },
+    #[error("error reading portable recording")]
+    PortableRecordingReadError {
+        #[source]
+        err: PortableRecordingReadError,
     },
     #[error("error reconstructing test list from archive")]
     TestListFromSummaryError {
@@ -588,6 +598,7 @@ impl ExpectedError {
             | Self::ShowTestGroupsError { .. }
             | Self::InvalidMessageFormatVersion { .. }
             | Self::DebugExtractReadError { .. }
+            | Self::DebugExtractRecordingError { .. }
             | Self::CargoMessageFormatError { .. } => NextestExitCode::SETUP_ERROR,
             Self::ConfigParseError { err } => {
                 // Experimental features not being enabled are their own error.
@@ -635,8 +646,10 @@ impl ExpectedError {
             | Self::RecordSessionSetupError { .. }
             | Self::RunIdResolutionError { .. }
             | Self::RecordReadError { .. }
-            | Self::UnsupportedStoreFormatVersion { .. }
+            | Self::StoreVersionIncompatible { .. }
             | Self::TestListFromSummaryError { .. } => NextestExitCode::SETUP_ERROR,
+            Self::PortableRecordingError { .. } => NextestExitCode::ARCHIVE_CREATION_FAILED,
+            Self::PortableRecordingReadError { .. } => NextestExitCode::SETUP_ERROR,
             Self::WriteError { .. } => NextestExitCode::WRITE_OUTPUT_ERROR,
             Self::FiltersetParseError { .. } => NextestExitCode::INVALID_FILTERSET,
         }
@@ -1053,7 +1066,7 @@ impl ExpectedError {
                     count.style(styles.bold),
                     // We only support reruns for test mode.
                     plural::tests_str(NextestRunMode::Test, *count),
-                    plural::remains_str(*count),
+                    plural::remain_str(*count),
                 );
                 // Advice to run `cargo nextest run -R latest` might not always
                 // be complete due to the expanded build scope or disappearing
@@ -1262,6 +1275,10 @@ impl ExpectedError {
                 error!("error writing {format} output");
                 Some(err as &dyn Error)
             }
+            Self::DebugExtractRecordingError { message } => {
+                error!("{message}");
+                None
+            }
             Self::RunIdResolutionError { err } => {
                 match &err {
                     RunIdResolutionError::NotFound { prefix } => {
@@ -1321,17 +1338,22 @@ impl ExpectedError {
                 error!("error reading recorded run");
                 Some(err as &dyn Error)
             }
-            Self::UnsupportedStoreFormatVersion {
+            Self::PortableRecordingError { err } => {
+                error!("error creating portable recording");
+                Some(err as &dyn Error)
+            }
+            Self::PortableRecordingReadError { err } => {
+                error!("error reading portable recording");
+                Some(err as &dyn Error)
+            }
+            Self::StoreVersionIncompatible {
                 run_id,
-                found,
-                supported,
+                incompatibility,
             } => {
                 error!(
-                    "run {} has unsupported store format version {} \
-                     (this nextest supports version {})",
+                    "run {} has incompatible store format: {}",
                     run_id.style(styles.bold),
-                    found.style(styles.bold),
-                    supported.style(styles.bold),
+                    incompatibility,
                 );
                 None
             }
