@@ -14,24 +14,32 @@ use crate::{
         scripts::{ScriptId, SetupScriptConfig},
     },
     input::{InputEvent, InputHandler},
-    list::{TestInstance, TestInstanceId, TestList},
+    list::{OwnedTestInstanceId, TestInstance, TestInstanceId, TestInstanceIdKey, TestList},
     reporter::events::{
-        CancelReason, ExecuteStatus, ExecutionStatuses, FinalRunStats, InfoResponse, ReporterEvent,
+        CancelReason, ChildExecutionOutputDescription, ExecuteStatus, ExecutionResultDescription,
+        ExecutionStatuses, FailureDescription, FinalRunStats, InfoResponse, ReporterEvent,
         RunFinishedStats, RunStats, StressIndex, StressProgress, StressRunStats, TestEvent,
-        TestEventKind,
+        TestEventKind, TestsNotSeen,
     },
     runner::{ExecutorEvent, RunUnitQuery, SignalRequest, StressCondition, StressCount},
     signal::{
         JobControlEvent, ShutdownEvent, ShutdownSignalEvent, SignalEvent, SignalHandler,
         SignalInfoEvent,
     },
+    test_output::ChildSingleOutput,
     time::StopwatchStart,
 };
 use chrono::Local;
 use debug_ignore::DebugIgnore;
 use futures::future::{Fuse, FusedFuture};
+use nextest_metadata::MismatchReason;
 use quick_junit::ReportUuid;
-use std::{collections::BTreeMap, env, pin::Pin, time::Duration};
+use std::{
+    collections::{BTreeMap, BTreeSet},
+    env, mem,
+    pin::Pin,
+    time::Duration,
+};
 use tokio::{
     sync::{
         mpsc::{UnboundedReceiver, UnboundedSender, unbounded_channel},
@@ -61,6 +69,7 @@ pub(super) struct DispatcherContext<'a, F> {
     signal_count: Option<SignalCount>,
     stress_cx: DispatcherStressContext,
     tick_interval: Duration,
+    rerun_cx: DispatcherRerunContext,
     #[cfg(test)]
     disable_signal_3_times_panic: bool,
 }
@@ -79,6 +88,7 @@ where
         max_fail: MaxFail,
         global_timeout: Duration,
         stress_condition: Option<StressCondition>,
+        expected_outstanding: Option<BTreeSet<OwnedTestInstanceId>>,
     ) -> Self {
         // Tick every 50ms by default.
         let tick_interval_ms = env::var("NEXTEST_PROGRESS_TICK_INTERVAL_MS")
@@ -102,6 +112,7 @@ where
             signal_count: None,
             stress_cx: DispatcherStressContext::new(stress_condition),
             tick_interval: Duration::from_millis(tick_interval_ms),
+            rerun_cx: DispatcherRerunContext::new(expected_outstanding),
             #[cfg(test)]
             disable_signal_3_times_panic: false,
         }
@@ -361,7 +372,7 @@ where
     pub(super) fn run_started(&mut self, test_list: &'a TestList, test_threads: usize) {
         let (stress_count, stress_infinite, stress_duration_nanos) =
             match self.stress_cx.condition() {
-                Some(StressCondition::Count(StressCount::Count(count))) => {
+                Some(StressCondition::Count(StressCount::Count { count })) => {
                     (Some(count.get()), false, None)
                 }
                 Some(StressCondition::Count(StressCount::Infinite)) => (None, true, None),
@@ -409,7 +420,7 @@ where
                 elapsed: _,
             } => {
                 let total = match total {
-                    StressCount::Count(n) => Some(n.get()),
+                    StressCount::Count { count } => Some(count.get()),
                     StressCount::Infinite => None,
                 };
                 (*completed, total)
@@ -448,7 +459,7 @@ where
                 elapsed: _,
             } => {
                 let total = match total {
-                    StressCount::Count(n) => Some(n.get()),
+                    StressCount::Count { count } => Some(count.get()),
                     StressCount::Infinite => None,
                 };
                 (*completed - 1, total)
@@ -551,7 +562,7 @@ where
                     total,
                     script_id,
                     program,
-                    args: &config.command.args,
+                    args: config.command.args.clone(),
                     no_capture: config.no_capture(),
                 })
             }
@@ -580,7 +591,7 @@ where
                     stress_index,
                     script_id,
                     program,
-                    args: &config.command.args,
+                    args: config.command.args.clone(),
                     elapsed,
                     will_terminate: will_terminate.is_some(),
                 })
@@ -599,20 +610,20 @@ where
 
                 // Fire the setup-script-done probe, extracting the exit code
                 // from the result if available.
-                let exit_code = match status.result {
-                    crate::reporter::events::ExecutionResult::Fail {
-                        failure_status: crate::reporter::events::FailureStatus::ExitCode(code),
+                let exit_code = match &status.result {
+                    ExecutionResultDescription::Fail {
+                        failure: FailureDescription::ExitCode { code },
                         ..
-                    } => Some(code),
+                    } => Some(*code),
                     _ => None,
                 };
 
-                // Extract stdout and stderr lengths from the output
+                // Extract stdout and stderr lengths from the output.
                 let (stdout_len, stderr_len) = match &status.output {
-                    crate::test_output::ChildExecutionOutput::Output { output, .. } => {
+                    ChildExecutionOutputDescription::Output { output, .. } => {
                         output.stdout_stderr_len()
                     }
-                    crate::test_output::ChildExecutionOutput::StartError(_) => (None, None),
+                    ChildExecutionOutputDescription::StartError(_) => (None, None),
                 };
 
                 crate::fire_usdt!(UsdtSetupScriptDone {
@@ -640,7 +651,7 @@ where
                     total,
                     script_id,
                     program,
-                    args: &config.command.args,
+                    args: config.command.args.clone(),
                     no_capture: config.no_capture(),
                     junit_store_success_output: config.junit.store_success_output,
                     junit_store_failure_output: config.junit.store_failure_output,
@@ -822,7 +833,12 @@ where
                 test_instance,
                 reason,
             }) => {
-                self.run_stats.skipped += 1;
+                // If the mismatch reason is that this test isn't a benchmark,
+                // we don't display it in the skip counts (but still keep track
+                // of it internally).
+                if !matches!(reason, MismatchReason::NotBenchmark) {
+                    self.run_stats.skipped += 1;
+                }
                 self.callback_none_response(TestEventKind::TestSkipped {
                     stress_index,
                     test_instance: test_instance.id(),
@@ -883,6 +899,9 @@ where
         instance: TestInstance<'a>,
         req_tx: UnboundedSender<RunUnitRequest<'a>>,
     ) {
+        // Track this test as seen for rerun tracking.
+        self.rerun_cx.mark_seen(instance.id());
+
         let prev = self.running_tests.insert(
             instance.id(),
             ContextTestInstance {
@@ -905,8 +924,8 @@ where
     fn finish_test(
         &mut self,
         key: TestInstanceId<'a>,
-        last_run_status: ExecuteStatus,
-    ) -> ExecutionStatuses {
+        last_run_status: ExecuteStatus<ChildSingleOutput>,
+    ) -> ExecutionStatuses<ChildSingleOutput> {
         self.running_tests
             .remove(&key)
             .unwrap_or_else(|| {
@@ -1114,6 +1133,9 @@ where
             stress_failed,
         });
 
+        let rerun_cx = mem::replace(&mut self.rerun_cx, DispatcherRerunContext::InitialRun);
+        let tests_not_seen = rerun_cx.into_tests_not_seen();
+
         self.basic_callback(TestEventKind::RunFinished {
             start_time: stopwatch_end.start_time.fixed_offset(),
             run_id: self.run_id,
@@ -1122,7 +1144,8 @@ where
                 || RunFinishedStats::Single(self.run_stats),
                 RunFinishedStats::Stress,
             ),
-        })
+            outstanding_not_seen: tests_not_seen,
+        });
     }
 
     pub(super) fn run_stats(&self) -> RunStats {
@@ -1142,6 +1165,56 @@ enum DispatcherStressContext {
     },
 }
 
+/// Context for rerun tracking.
+///
+/// This enum tracks whether the current run is an initial run or a rerun, and
+/// maintains the necessary state for computing which expected tests were not
+/// seen during a rerun.
+#[derive(Clone, Debug)]
+enum DispatcherRerunContext {
+    /// This is an initial run, not a rerun. No tracking needed.
+    InitialRun,
+    /// This is a rerun of a previous run.
+    ///
+    /// Contains the set of tests expected to run. As tests are seen, they are
+    /// removed from this set. At the end of the run, any remaining tests are
+    /// reported as "not seen".
+    Rerun(BTreeSet<OwnedTestInstanceId>),
+}
+
+impl DispatcherRerunContext {
+    fn new(expected_outstanding: Option<BTreeSet<OwnedTestInstanceId>>) -> Self {
+        match expected_outstanding {
+            Some(expected) => Self::Rerun(expected),
+            None => Self::InitialRun,
+        }
+    }
+
+    /// Marks a test as seen during this run.
+    fn mark_seen(&mut self, id: TestInstanceId<'_>) {
+        if let Self::Rerun(expected) = self {
+            expected.remove(&id as &dyn TestInstanceIdKey);
+        }
+    }
+
+    /// Returns tests not seen, if this is a rerun and some tests were not seen.
+    ///
+    /// Returns `None` if this is an initial run.
+    fn into_tests_not_seen(self) -> Option<TestsNotSeen> {
+        let Self::Rerun(not_seen) = self else {
+            return None;
+        };
+
+        let total_not_seen = not_seen.len();
+        const MAX_DISPLAY: usize = 8;
+        let sample: Vec<_> = not_seen.into_iter().take(MAX_DISPLAY).collect();
+        Some(TestsNotSeen {
+            not_seen: sample,
+            total_not_seen,
+        })
+    }
+}
+
 impl DispatcherStressContext {
     fn new(condition: Option<StressCondition>) -> Self {
         if let Some(condition) = condition {
@@ -1201,7 +1274,7 @@ impl DispatcherStressContext {
                 // completed runs.
                 let current = *completed;
                 let total = match condition {
-                    StressCondition::Count(StressCount::Count(total)) => Some(*total),
+                    StressCondition::Count(StressCount::Count { count }) => Some(*count),
                     StressCondition::Count(StressCount::Infinite)
                     | StressCondition::Duration(_) => None,
                 };
@@ -1229,7 +1302,7 @@ impl DispatcherStressContext {
                         // TODO: We should figure out whether to terminate the
                         // test run based on this.
                     }
-                    FinalRunStats::Failed(_) => {
+                    FinalRunStats::Failed { .. } => {
                         *failed += 1;
                     }
                     FinalRunStats::Cancelled { .. } => {
@@ -1324,16 +1397,19 @@ struct ContextTestInstance<'a> {
     // Store the instance primarily for debugging.
     #[expect(dead_code)]
     instance: TestInstance<'a>,
-    past_attempts: Vec<ExecuteStatus>,
+    past_attempts: Vec<ExecuteStatus<ChildSingleOutput>>,
     req_tx: UnboundedSender<RunUnitRequest<'a>>,
 }
 
 impl ContextTestInstance<'_> {
-    fn attempt_failed_will_retry(&mut self, run_status: ExecuteStatus) {
+    fn attempt_failed_will_retry(&mut self, run_status: ExecuteStatus<ChildSingleOutput>) {
         self.past_attempts.push(run_status);
     }
 
-    fn finish(self, last_run_status: ExecuteStatus) -> ExecutionStatuses {
+    fn finish(
+        self,
+        last_run_status: ExecuteStatus<ChildSingleOutput>,
+    ) -> ExecutionStatuses<ChildSingleOutput> {
         let mut attempts = self.past_attempts;
         attempts.push(last_run_status);
         ExecutionStatuses::new(attempts)
@@ -1427,7 +1503,8 @@ mod tests {
             0,
             MaxFail::All,
             crate::time::far_future_duration(),
-            None,
+            None, // stress_condition
+            None, // expected_outstanding
         );
 
         cx.disable_signal_3_times_panic = true;
