@@ -9,10 +9,12 @@ use crate::{
     errors::TestFilterBuilderError,
     list::RustTestArtifact,
     partition::{Partitioner, PartitionerBuilder},
+    record::ComputedRerunInfo,
+    run_mode::NextestRunMode,
 };
 use aho_corasick::AhoCorasick;
 use nextest_filtering::{EvalContext, Filterset, TestQuery};
-use nextest_metadata::{FilterMatch, MismatchReason};
+use nextest_metadata::{FilterMatch, MismatchReason, RustTestKind, TestCaseName};
 use std::{collections::HashSet, fmt, mem};
 
 /// Whether to run ignored tests.
@@ -101,8 +103,10 @@ impl BinaryFilter {
 }
 
 /// A builder for `TestFilter` instances.
-#[derive(Clone, Debug, Eq, PartialEq)]
+#[derive(Clone, Debug)]
 pub struct TestFilterBuilder {
+    mode: NextestRunMode,
+    rerun_info: Option<ComputedRerunInfo>,
     run_ignored: RunIgnored,
     partitioner_builder: Option<PartitionerBuilder>,
     patterns: ResolvedFilterPatterns,
@@ -294,12 +298,13 @@ impl TestFilterPatterns {
     }
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, Default)]
 enum ResolvedFilterPatterns {
     /// Match all tests.
     ///
     /// This is mostly for convenience -- it's equivalent to `SkipOnly` with an empty set of skip
     /// patterns.
+    #[default]
     All,
 
     /// Match all tests except those that match the skip patterns.
@@ -320,14 +325,9 @@ enum ResolvedFilterPatterns {
     },
 }
 
-impl Default for ResolvedFilterPatterns {
-    fn default() -> Self {
-        Self::All
-    }
-}
-
 impl ResolvedFilterPatterns {
-    fn name_match(&self, test_name: &str) -> FilterNameMatch {
+    fn name_match(&self, test_name: &TestCaseName) -> FilterNameMatch {
+        let test_name = test_name.as_str();
         match self {
             Self::All => FilterNameMatch::MatchEmptyPatterns,
             Self::SkipOnly {
@@ -427,6 +427,7 @@ impl TestFilterBuilder {
     ///
     /// If an empty slice is passed, the test filter matches all possible test names.
     pub fn new(
+        mode: NextestRunMode,
         run_ignored: RunIgnored,
         partitioner_builder: Option<PartitionerBuilder>,
         patterns: TestFilterPatterns,
@@ -437,6 +438,8 @@ impl TestFilterBuilder {
         let binary_filter = BinaryFilter::new(exprs);
 
         Ok(Self {
+            mode,
+            rerun_info: None,
             run_ignored,
             partitioner_builder,
             patterns,
@@ -459,9 +462,11 @@ impl TestFilterBuilder {
     }
 
     /// Creates a new `TestFilterBuilder` that matches the default set of tests.
-    pub fn default_set(run_ignored: RunIgnored) -> Self {
+    pub fn default_set(mode: NextestRunMode, run_ignored: RunIgnored) -> Self {
         let binary_filter = BinaryFilter::new(Vec::new());
         Self {
+            mode,
+            rerun_info: None,
             run_ignored,
             partitioner_builder: None,
             patterns: ResolvedFilterPatterns::default(),
@@ -469,6 +474,21 @@ impl TestFilterBuilder {
         }
     }
 
+    /// Set the list of outstanding tests, if this is a rerun.
+    pub fn set_outstanding_tests(&mut self, rerun_info: ComputedRerunInfo) {
+        self.rerun_info = Some(rerun_info);
+    }
+
+    /// Returns the nextest execution mode.
+    pub fn mode(&self) -> NextestRunMode {
+        self.mode
+    }
+
+    /// Compares the patterns between two `TestFilterBuilder`s.
+    pub fn patterns_eq(&self, other: &Self) -> bool {
+        self.patterns == other.patterns
+    }
+
     /// Creates a new test filter scoped to a single binary.
     ///
     /// This test filter may be stateful.
@@ -482,6 +502,11 @@ impl TestFilterBuilder {
             partitioner,
         }
     }
+
+    /// Consumes self, returning the underlying [`ComputedRerunInfo`] if any.
+    pub fn into_rerun_info(self) -> Option<ComputedRerunInfo> {
+        self.rerun_info
+    }
 }
 
 /// Whether a binary matched filters and should be run to obtain the list of tests within.
@@ -590,58 +615,127 @@ impl TestFilter<'_> {
     pub fn filter_match(
         &mut self,
         test_binary: &RustTestArtifact<'_>,
-        test_name: &str,
+        test_name: &TestCaseName,
+        test_kind: &RustTestKind,
         ecx: &EvalContext<'_>,
         bound: FilterBound,
         ignored: bool,
     ) -> FilterMatch {
-        self.filter_ignored_mismatch(ignored)
-            .or_else(|| {
-                // ---
-                // NOTE
-                // ---
-                //
-                // Previously, if either expression OR string filters matched, we'd run the tests.
-                // The current (stable) implementation is that *both* the expression AND the string
-                // filters should match.
-                //
-                // This is because we try and skip running test binaries which don't match
-                // expression filters. So for example:
-                //
-                //     cargo nextest run -E 'binary(foo)' test_bar
-                //
-                // would not even get to the point of enumerating the tests not in binary(foo), thus
-                // not running any test_bars in the workspace. But, with the OR semantics:
-                //
-                //     cargo nextest run -E 'binary(foo) or test(test_foo)' test_bar
-                //
-                // would run all the test_bars in the repo. This is inconsistent, so nextest must
-                // use AND semantics.
-                use FilterNameMatch::*;
-                match (
-                    self.filter_name_match(test_name),
-                    self.filter_expression_match(test_binary, test_name, ecx, bound),
-                ) {
-                    // Tests must be accepted by both expressions and filters.
-                    (
-                        MatchEmptyPatterns | MatchWithPatterns,
-                        MatchEmptyPatterns | MatchWithPatterns,
-                    ) => None,
-                    // If rejected by at least one of the filtering strategies, the test is
-                    // rejected. Note we use the _name_ mismatch reason first. That's because
-                    // expression-based matches can also match against the default set. If a test
-                    // fails both name and expression matches, then the name reason is more directly
-                    // relevant.
-                    (Mismatch(reason), _) | (_, Mismatch(reason)) => {
-                        Some(FilterMatch::Mismatch { reason })
-                    }
+        // Handle benchmark mismatches first.
+        if let Some(mismatch) = self.filter_benchmark_mismatch(test_kind) {
+            return mismatch;
+        }
+
+        // Check if this test already passed in a prior rerun.
+        //
+        // RerunAlreadyPassed is a high-order bit: if a test passed in a prior
+        // rerun, we shouldn't run it again, regardless of other filter results.
+        // However, we must still go through the motions of checking all other
+        // filters, particularly for counted partitioning, to maintain
+        // consistent bucketing across reruns.
+        //
+        // Note that we don't support reruns with benchmarks yet (probably
+        // ever?), so NotABenchmark and RerunAlreadyPassed are mutually
+        // exclusive.
+        if self.is_rerun_already_passed(test_binary, test_name) {
+            // Run through the base filter to maintain partition counts.
+            let _ = self.filter_match_base(test_binary, test_name, ecx, bound, ignored);
+            return FilterMatch::Mismatch {
+                reason: MismatchReason::RerunAlreadyPassed,
+            };
+        }
+
+        self.filter_match_base(test_binary, test_name, ecx, bound, ignored)
+    }
+
+    /// Core filter matching logic, used by `filter_match`.
+    fn filter_match_base(
+        &mut self,
+        test_binary: &RustTestArtifact<'_>,
+        test_name: &TestCaseName,
+        ecx: &EvalContext<'_>,
+        bound: FilterBound,
+        ignored: bool,
+    ) -> FilterMatch {
+        if let Some(mismatch) = self.filter_ignored_mismatch(ignored) {
+            return mismatch;
+        }
+
+        {
+            // ---
+            // NOTE
+            // ---
+            //
+            // Previously, if either expression OR string filters matched, we'd run the tests.
+            // The current (stable) implementation is that *both* the expression AND the string
+            // filters should match.
+            //
+            // This is because we try and skip running test binaries which don't match
+            // expression filters. So for example:
+            //
+            //     cargo nextest run -E 'binary(foo)' test_bar
+            //
+            // would not even get to the point of enumerating the tests not in binary(foo), thus
+            // not running any test_bars in the workspace. But, with the OR semantics:
+            //
+            //     cargo nextest run -E 'binary(foo) or test(test_foo)' test_bar
+            //
+            // would run all the test_bars in the repo. This is inconsistent, so nextest must
+            // use AND semantics.
+            use FilterNameMatch::*;
+            match (
+                self.filter_name_match(test_name),
+                self.filter_expression_match(test_binary, test_name, ecx, bound),
+            ) {
+                // Tests must be accepted by both expressions and filters.
+                (
+                    MatchEmptyPatterns | MatchWithPatterns,
+                    MatchEmptyPatterns | MatchWithPatterns,
+                ) => {}
+                // If rejected by at least one of the filtering strategies, the test is
+                // rejected. Note we use the _name_ mismatch reason first. That's because
+                // expression-based matches can also match against the default set. If a test
+                // fails both name and expression matches, then the name reason is more directly
+                // relevant.
+                (Mismatch(reason), _) | (_, Mismatch(reason)) => {
+                    return FilterMatch::Mismatch { reason };
                 }
+            }
+        }
+
+        // Note that partition-based filtering MUST come after all other kinds
+        // of filtering, so that count-based bucketing applies after ignored,
+        // name, and expression matching. This also means that mutable count
+        // state must be maintained by the partitioner.
+        if let Some(mismatch) = self.filter_partition_mismatch(test_name) {
+            return mismatch;
+        }
+
+        FilterMatch::Matches
+    }
+
+    /// Returns true if this test already passed in a prior rerun.
+    fn is_rerun_already_passed(
+        &self,
+        test_binary: &RustTestArtifact<'_>,
+        test_name: &TestCaseName,
+    ) -> bool {
+        if let Some(rerun_info) = &self.builder.rerun_info
+            && let Some(suite) = rerun_info.test_suites.get(&test_binary.binary_id)
+        {
+            return suite.passing.contains(test_name);
+        }
+        false
+    }
+
+    fn filter_benchmark_mismatch(&self, test_kind: &RustTestKind) -> Option<FilterMatch> {
+        if self.builder.mode == NextestRunMode::Benchmark && test_kind != &RustTestKind::BENCH {
+            Some(FilterMatch::Mismatch {
+                reason: MismatchReason::NotBenchmark,
             })
-            // Note that partition-based filtering MUST come after all other kinds of filtering,
-            // so that count-based bucketing applies after ignored, name and expression matching.
-            // This also means that mutable count state must be maintained by the partitioner.
-            .or_else(|| self.filter_partition_mismatch(test_name))
-            .unwrap_or(FilterMatch::Matches)
+        } else {
+            None
+        }
     }
 
     fn filter_ignored_mismatch(&self, ignored: bool) -> Option<FilterMatch> {
@@ -665,14 +759,14 @@ impl TestFilter<'_> {
         None
     }
 
-    fn filter_name_match(&self, test_name: &str) -> FilterNameMatch {
+    fn filter_name_match(&self, test_name: &TestCaseName) -> FilterNameMatch {
         self.builder.patterns.name_match(test_name)
     }
 
     fn filter_expression_match(
         &self,
         test_binary: &RustTestArtifact<'_>,
-        test_name: &str,
+        test_name: &TestCaseName,
         ecx: &EvalContext<'_>,
         bound: FilterBound,
     ) -> FilterNameMatch {
@@ -704,9 +798,9 @@ impl TestFilter<'_> {
         }
     }
 
-    fn filter_partition_mismatch(&mut self, test_name: &str) -> Option<FilterMatch> {
+    fn filter_partition_mismatch(&mut self, test_name: &TestCaseName) -> Option<FilterMatch> {
         let partition_match = match &mut self.partitioner {
-            Some(partitioner) => partitioner.test_matches(test_name),
+            Some(partitioner) => partitioner.test_matches(test_name.as_str()),
             None => true,
         };
         if partition_match {
@@ -748,10 +842,17 @@ mod tests {
     #[proptest(cases = 50)]
     fn proptest_empty(#[strategy(vec(any::<String>(), 0..16))] test_names: Vec<String>) {
         let patterns = TestFilterPatterns::default();
-        let test_filter =
-            TestFilterBuilder::new(RunIgnored::Default, None, patterns, Vec::new()).unwrap();
+        let test_filter = TestFilterBuilder::new(
+            NextestRunMode::Test,
+            RunIgnored::Default,
+            None,
+            patterns,
+            Vec::new(),
+        )
+        .unwrap();
         let single_filter = test_filter.build();
         for test_name in test_names {
+            let test_name = TestCaseName::new(&test_name);
             prop_assert!(single_filter.filter_name_match(&test_name).is_match());
         }
     }
@@ -761,11 +862,18 @@ mod tests {
     fn proptest_exact(#[strategy(vec(any::<String>(), 0..16))] test_names: Vec<String>) {
         // Test with the default matcher.
         let patterns = TestFilterPatterns::new(test_names.clone());
-        let test_filter =
-            TestFilterBuilder::new(RunIgnored::Default, None, patterns, Vec::new()).unwrap();
+        let test_filter = TestFilterBuilder::new(
+            NextestRunMode::Test,
+            RunIgnored::Default,
+            None,
+            patterns,
+            Vec::new(),
+        )
+        .unwrap();
         let single_filter = test_filter.build();
         for test_name in &test_names {
-            prop_assert!(single_filter.filter_name_match(test_name).is_match());
+            let test_name = TestCaseName::new(test_name);
+            prop_assert!(single_filter.filter_name_match(&test_name).is_match());
         }
 
         // Test with the exact matcher.
@@ -773,11 +881,18 @@ mod tests {
         for test_name in &test_names {
             patterns.add_exact_pattern(test_name.clone());
         }
-        let test_filter =
-            TestFilterBuilder::new(RunIgnored::Default, None, patterns, Vec::new()).unwrap();
+        let test_filter = TestFilterBuilder::new(
+            NextestRunMode::Test,
+            RunIgnored::Default,
+            None,
+            patterns,
+            Vec::new(),
+        )
+        .unwrap();
         let single_filter = test_filter.build();
         for test_name in &test_names {
-            prop_assert!(single_filter.filter_name_match(test_name).is_match());
+            let test_name = TestCaseName::new(test_name);
+            prop_assert!(single_filter.filter_name_match(&test_name).is_match());
         }
     }
 
@@ -793,10 +908,17 @@ mod tests {
             patterns.add_substring_pattern(substring);
         }
 
-        let test_filter =
-            TestFilterBuilder::new(RunIgnored::Default, None, patterns, Vec::new()).unwrap();
+        let test_filter = TestFilterBuilder::new(
+            NextestRunMode::Test,
+            RunIgnored::Default,
+            None,
+            patterns,
+            Vec::new(),
+        )
+        .unwrap();
         let single_filter = test_filter.build();
         for test_name in test_names {
+            let test_name = TestCaseName::new(&test_name);
             prop_assert!(single_filter.filter_name_match(&test_name).is_match());
         }
     }
@@ -807,12 +929,23 @@ mod tests {
         prop_assume!(!substring.is_empty() && !prefix.is_empty() && !suffix.is_empty());
         let pattern = prefix + &substring + &suffix;
         let patterns = TestFilterPatterns::new(vec![pattern]);
-        let test_filter =
-            TestFilterBuilder::new(RunIgnored::Default, None, patterns, Vec::new()).unwrap();
+        let test_filter = TestFilterBuilder::new(
+            NextestRunMode::Test,
+            RunIgnored::Default,
+            None,
+            patterns,
+            Vec::new(),
+        )
+        .unwrap();
         let single_filter = test_filter.build();
+        let substring = TestCaseName::new(&substring);
         prop_assert!(!single_filter.filter_name_match(&substring).is_match());
     }
 
+    fn test_name(s: &str) -> TestCaseName {
+        TestCaseName::new(s)
+    }
+
     #[test]
     fn pattern_examples() {
         let mut patterns = TestFilterPatterns::new(vec!["foo".to_string()]);
@@ -825,57 +958,57 @@ mod tests {
 
         // Test substring matches.
         assert_eq!(
-            resolved.name_match("foo"),
+            resolved.name_match(&test_name("foo")),
             FilterNameMatch::MatchWithPatterns,
         );
         assert_eq!(
-            resolved.name_match("1foo2"),
+            resolved.name_match(&test_name("1foo2")),
             FilterNameMatch::MatchWithPatterns,
         );
         assert_eq!(
-            resolved.name_match("bar"),
+            resolved.name_match(&test_name("bar")),
             FilterNameMatch::MatchWithPatterns,
         );
         assert_eq!(
-            resolved.name_match("x_bar_y"),
+            resolved.name_match(&test_name("x_bar_y")),
             FilterNameMatch::MatchWithPatterns,
         );
 
         // Test exact matches.
         assert_eq!(
-            resolved.name_match("baz"),
+            resolved.name_match(&test_name("baz")),
             FilterNameMatch::MatchWithPatterns,
         );
         assert_eq!(
-            resolved.name_match("abazb"),
+            resolved.name_match(&test_name("abazb")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
 
         // Both substring and exact matches.
         assert_eq!(
-            resolved.name_match("bazfoo"),
+            resolved.name_match(&test_name("bazfoo")),
             FilterNameMatch::MatchWithPatterns,
         );
 
         // Skip patterns.
         assert_eq!(
-            resolved.name_match("quux"),
+            resolved.name_match(&test_name("quux")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
         assert_eq!(
-            resolved.name_match("1quux2"),
+            resolved.name_match(&test_name("1quux2")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
 
         // Skip and substring patterns.
         assert_eq!(
-            resolved.name_match("quuxbar"),
+            resolved.name_match(&test_name("quuxbar")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
 
         // Skip-exact patterns.
         assert_eq!(
-            resolved.name_match("quuz"),
+            resolved.name_match(&test_name("quuz")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
 
@@ -883,7 +1016,7 @@ mod tests {
         patterns.add_skip_pattern("baz".to_string());
         let resolved = patterns.resolve().unwrap();
         assert_eq!(
-            resolved.name_match("quuxbaz"),
+            resolved.name_match(&test_name("quuxbaz")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
     }
@@ -899,35 +1032,35 @@ mod tests {
 
         // Test substring matches.
         assert_eq!(
-            resolved.name_match("foo"),
+            resolved.name_match(&test_name("foo")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
         assert_eq!(
-            resolved.name_match("1foo2"),
+            resolved.name_match(&test_name("1foo2")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
         assert_eq!(
-            resolved.name_match("bar"),
+            resolved.name_match(&test_name("bar")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
         assert_eq!(
-            resolved.name_match("x_bar_y"),
+            resolved.name_match(&test_name("x_bar_y")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
 
         // Test exact matches.
         assert_eq!(
-            resolved.name_match("baz"),
+            resolved.name_match(&test_name("baz")),
             FilterNameMatch::Mismatch(MismatchReason::String),
         );
         assert_eq!(
-            resolved.name_match("abazb"),
+            resolved.name_match(&test_name("abazb")),
             FilterNameMatch::MatchWithPatterns,
         );
 
         // Anything that doesn't match the skip filter should match.
         assert_eq!(
-            resolved.name_match("quux"),
+            resolved.name_match(&test_name("quux")),
             FilterNameMatch::MatchWithPatterns,
         );
     }
@@ -940,27 +1073,27 @@ mod tests {
 
         // Anything matches.
         assert_eq!(
-            resolved.name_match("foo"),
+            resolved.name_match(&test_name("foo")),
             FilterNameMatch::MatchEmptyPatterns,
         );
         assert_eq!(
-            resolved.name_match("1foo2"),
+            resolved.name_match(&test_name("1foo2")),
             FilterNameMatch::MatchEmptyPatterns,
         );
         assert_eq!(
-            resolved.name_match("bar"),
+            resolved.name_match(&test_name("bar")),
             FilterNameMatch::MatchEmptyPatterns,
         );
         assert_eq!(
-            resolved.name_match("x_bar_y"),
+            resolved.name_match(&test_name("x_bar_y")),
             FilterNameMatch::MatchEmptyPatterns,
         );
         assert_eq!(
-            resolved.name_match("baz"),
+            resolved.name_match(&test_name("baz")),
             FilterNameMatch::MatchEmptyPatterns,
         );
         assert_eq!(
-            resolved.name_match("abazb"),
+            resolved.name_match(&test_name("abazb")),
             FilterNameMatch::MatchEmptyPatterns,
         );
     }
